/*
 * TOBY -- An abstract interpreter engine and system for learning.
 * Copyright (C) 1999  Ryan C. Gordon.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "platform/lexers/xmllexer/XMLLexer.h"


#include "io/FileReader.h"  // !!! TEMP.

TobyCollection *XMLLexer::elements = NULL;


    // (tag) is the XML tag in question.
    // (func) should return a pointer to an appropriate LexerRules, built
    //  based on the details of (node). The passed node should be correct
    //  (only <element> nodes should be passed to ElementRulesXML), but it
    //  couldn't hurt to assert this fact. The rules in their entirety need
    //  to be be built. That is, an ElementRulesXML class should build all
    //  it's child rules, too, preferrably by passing their nodes through
    //  XMLLexer::buildRules(). This makes the buildRules() method more
    //  complex, but keeps the parent class as lean and fast as possible,
    //  so all the overhead goes away if we generate the rules as C++
    //  with outputConstructor(), etc.
typedef struct
{
    const char *tag;
    LexerRules *(*func)(XMLNode *node);
} RulesTagHandlers;

static RulesTagHandlers handlers[] =
{
    { "reqelement",           ReqElementRulesXML::buildRules },
    { "reqword",              ReqWordRulesXML::buildRules },
    { "reqnumber",            ReqNumberRulesXML::buildRules },
    { "reqeof",               ReqEOFRulesXML::buildRules },
    { "reqnewline",           ReqNewlineRulesXML::buildRules },
    { "reqwordchars",         ReqWordCharsRulesXML::buildRules },
    { "reqchar",              ReqCharRulesXML::buildRules },
    { "reqliteralstring",     ReqLiteralStringRulesXML::buildRules },
    { "reqwhitespace",        ReqWhitespaceRulesXML::buildRules },
    { "reqmultilinecomment",  ReqMultiLineCommentRulesXML::buildRules },
    { "reqsinglelinecomment", ReqSingleLineCommentRulesXML::buildRules },
    { "pickone",              PickOneRulesXML::buildRules },
    { "element",              ElementRulesXML::buildRules },
    { "repeat",               RepeatRulesXML::buildRulesRepeatTag },
    { "optional",             RepeatRulesXML::buildRulesOptionalTag },
    { "tokenizer",            TokenizerRulesXML::buildRules },
    { "language",             LanguageRulesXML::buildRules },
    { NULL,                   NULL }
};


LexerRules *XMLLexer::buildRules(XMLNode *node)
{
    const char *tag = node->getTag();

    for (int i = 0; handlers[i].tag != NULL; i++)
    {
        if (strcmp(tag, handlers[i].tag) == 0)
            return(handlers[i].func(node));
    } // for

    assert(false);
    return(NULL);
} // XMLLexer::buildRules


LexerRules **XMLLexer::buildBasicChildRules(XMLNode *node, size_t *count)
{
    TobyCollection *children = node->getChildren();
    size_t max = children->getSize();

    if (max <= 0)
    {
        *count = 0;
        return(NULL);
    } // if

    LexerRules **rulesList = new LexerRules *[max];
    for (size_t i = 0; i < max; i++)
    {
        XMLNode *kid = (XMLNode *) children->elementAt(i);
        rulesList[i] = XMLLexer::buildRules(kid);
        if (rulesList[i] == NULL)
        {
            for (size_t j = 0; j < i; j++)
                delete rulesList[j];
            delete[] rulesList;
            return(NULL);
        } // if
    } // for

    *count = max;
    return(rulesList);
} // XMLLexer::buildBasicChildRules


LexerRules **XMLLexer::buildLanguages(XMLTree *tree)
{
    XMLNode *root = tree->getRootNode();
    TobyCollection *children = root->getChildren();
    size_t max = children->getSize();
    LexerRules **retval = new LexerRules *[max + 1];
    size_t langCount = 0;

    XMLLexer::elements = new TobyCollection();

    for (size_t i = 0; i < max; i++)
    {
        XMLNode *node = (XMLNode *) children->elementAt(i);
        if (strcmp(node->getTag(), "language") == 0)
        {
            retval[langCount] = LanguageRulesXML::buildRules(node);
            if (retval[langCount] != NULL)
            {
                if (retval[langCount]->resolve(retval[langCount]) == false)
                    delete retval[langCount];
                else
                    langCount++;
            } // if

            elements->clear();  // Elements are only valid in their language.
        } // if
    } // for

    retval[langCount] = NULL;

    delete XMLLexer::elements;
    XMLLexer::elements = NULL;

    return(retval);
} // XMLLexer::buildLanguages


Lexer *Lexer::getInstance(void)
{
    return(new XMLLexer());
} // Lexer::getInstance


void XMLLexer::loadLanguages(void)
{
    // !!! Write a REAL implementation of this.

    XMLTree tree(new FileReader("grammars/toby.xml"));
    if (tree.parseXML())
    {
        LexerRules **langs = buildLanguages(&tree);
        for (size_t i = 0; langs[i] != NULL; i++)
            registerLanguage((LanguageRules *) langs[i]);
        delete[] langs;
    } // if
} // XMLLexer::loadLanguages


void XMLLexer::addElement(ElementRulesXML *elem)
{
    if (XMLLexer::elements != NULL)
        XMLLexer::elements->addElement(elem);
} // XMLLexer::addElement


ElementRulesXML *XMLLexer::getElementByName(const char *name)
{
    if (XMLLexer::elements != NULL)
    {
        size_t max = XMLLexer::elements->getSize();
        for (size_t i = 0; i < max; i++)
        {
            ElementRulesXML *elem;
            elem = (ElementRulesXML *) XMLLexer::elements->elementAt(i);
            if (strcmp(elem->getName(), name) == 0)
                return(elem);
        } // for
    } // if

    return(NULL);
} // XMLLexer::getElementByName

// end of XMLLexer.cpp ...

